/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <gnutls/x509.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>

#include "pwmd-error.h"
#include "mem.h"
#include "util-misc.h"
#include "util-string.h"
#include "common.h"
#include "tls.h"
#include "mutex.h"

#define WAIT_INTERVAL 50000
#define TEST_TIMEOUT(timeout, ret, start, rc)				\
  do {									\
    if (ret == GNUTLS_E_AGAIN)						\
      {									\
	time_t now = time (NULL);					\
	if (timeout && now - start >= timeout)				\
	  {								\
	    rc = gpg_error (GPG_ERR_ETIMEDOUT);				\
	    break;							\
	  }								\
      }									\
    if (ret != GNUTLS_E_INTERRUPTED)					\
      break;								\
    struct timeval tv = { 0, WAIT_INTERVAL };				\
    select (0, NULL, NULL, NULL, &tv);					\
  }									\
  while (0)

static char *
tls_fingerprint (gnutls_session_t ses)
{
  gnutls_x509_crt_t crt;
  const gnutls_datum_t *cert_list;
  unsigned count;
  unsigned char buf[20];
  size_t len = sizeof (buf);

  gnutls_x509_crt_init (&crt);
  if (!crt)
    {
      log_write ("%s(%i): %s(): %s", __FILE__, __LINE__, __FUNCTION__,
		 gnutls_strerror (GNUTLS_E_MEMORY_ERROR));
      return NULL;
    }

  cert_list = gnutls_certificate_get_peers (ses, &count);
  if (count)
    {
      gnutls_x509_crt_import (crt, &cert_list[0], GNUTLS_X509_FMT_DER);
      gnutls_x509_crt_get_fingerprint (crt, GNUTLS_DIG_SHA1, buf, &len);
    }

  gnutls_x509_crt_deinit (crt);
  return bin2hex (buf, len);
}

static int
verify_client_certificate (unsigned status)
{
  if (!status)
    return 0;

  if (status & GNUTLS_CERT_INVALID)
    log_write (_("client certificate is invalid"));

  if (status & GNUTLS_CERT_REVOKED)
    log_write (_("client certificate is revoked"));

  if (status & GNUTLS_CERT_SIGNER_NOT_FOUND)
    log_write (_("client certificate has no signer"));

  if (status & GNUTLS_CERT_SIGNER_NOT_CA)
    log_write (_("client certificate signer is not from CA"));

  if (status & GNUTLS_CERT_INSECURE_ALGORITHM)
    log_write (_("client certificate has insecure algorithm"));

  return GNUTLS_E_CERTIFICATE_ERROR;
}

struct tls_s *
tls_init (int fd, int timeout, const char *prio)
{
  struct tls_s *tls = xcalloc (1, sizeof (struct tls_s));
  int ret;
  unsigned status;
  const char *prio_error;
  gnutls_kx_algorithm_t kx;

  if (!tls)
    {
      log_write ("%s(%i): %s: %s", __FILE__, __LINE__, __FUNCTION__,
		 strerror (ENOMEM));
      return NULL;
    }

  ret = gnutls_init (&tls->ses, GNUTLS_SERVER);
  if (ret != GNUTLS_E_SUCCESS)
    goto fail;

  ret = gnutls_priority_set_direct (tls->ses, prio, &prio_error);
  if (ret != GNUTLS_E_SUCCESS)
    goto fail;

  ret = gnutls_credentials_set (tls->ses, GNUTLS_CRD_CERTIFICATE, x509_cred);
  if (ret != GNUTLS_E_SUCCESS)
    goto fail;

  gnutls_certificate_server_set_request (tls->ses, GNUTLS_CERT_REQUIRE);
  gnutls_transport_set_ptr (tls->ses, (gnutls_transport_ptr_t) fd);
  time_t start = time (NULL);
  do
    {
      gpg_error_t rc = 0;

      ret = gnutls_handshake (tls->ses);
      TEST_TIMEOUT (timeout, ret, start, rc);
      if (rc)
	goto fail;
    }
  while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);

  if (ret != GNUTLS_E_SUCCESS)
    goto fail;

  ret = gnutls_certificate_verify_peers2 (tls->ses, &status);
  if (ret)
    goto fail;

  kx = gnutls_kx_get (tls->ses);
  tls->fp = tls_fingerprint (tls->ses);
  log_write ("PROTO=%s CIPHER=%s MAC=%s KX=%s(%d) FP=%s",
	     gnutls_protocol_get_name (gnutls_protocol_get_version
				       (tls->ses)),
	     gnutls_cipher_get_name (gnutls_cipher_get (tls->ses)),
	     gnutls_mac_get_name (gnutls_mac_get (tls->ses)),
	     gnutls_kx_get_name (kx), gnutls_dh_get_prime_bits (tls->ses),
	     tls->fp ? tls->fp : "N/A");
  ret = verify_client_certificate (status);
  if (ret)
    goto fail;

  return tls;

fail:
  log_write ("%s", gnutls_strerror (ret));
  gnutls_deinit (tls->ses);
  xfree (tls->fp);
  xfree (tls);
  return NULL;
}

/* From the documentation. */
int
tls_get_params (gnutls_session_t ses, gnutls_params_type_t type,
		gnutls_params_st * st)
{
  if (type == GNUTLS_PARAMS_RSA_EXPORT)
    st->params.rsa_export = rsa_params;
  else if (type == GNUTLS_PARAMS_DH)
    st->params.dh = dh_params;
  else
    return -1;

  st->type = type;
  st->deinit = 0;
  return 0;
}

void
tls_log (int level, const char *msg)
{
  log_write ("TLS: %i: %s", level, msg);
}

ssize_t
tls_read_hook (assuan_context_t ctx, assuan_fd_t fd, void *data, size_t len)
{
  struct client_s *client = assuan_get_pointer (ctx);
  time_t start = time (NULL);
  ssize_t ret;

 rehandshake:
  do
    {
      gpg_error_t rc = 0;

      ret = gnutls_record_recv (client->thd->tls->ses, data, len);
      if (ret == GNUTLS_E_REHANDSHAKE)
	{
	  ret = gnutls_rehandshake (client->thd->tls->ses);
	  if (ret == GNUTLS_E_GOT_APPLICATION_DATA)
	    goto rehandshake;
	  else if (ret != GNUTLS_E_SUCCESS)
	    {
	      log_write ("%s", gnutls_strerror (ret));
	      ret = 0;
	      break;
	    }

	  do
	    {
	      ret = gnutls_handshake (client->thd->tls->ses);
	      rc = 0;
	      TEST_TIMEOUT (client->thd->timeout, ret, start, rc);
	      if (rc)
		{
		  log_write ("%s", gnutls_strerror (ret));
		  ret = 0;
		  close (client->thd->fd);
		  client->thd->fd = -1;
		  break;
		}
	    }
	  while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);

	  if (rc)
	    break;

	  continue;
	}

      TEST_TIMEOUT (client->thd->timeout, ret, start, rc);
      if (rc)
	{
	  errno = ETIMEDOUT;
	  close (client->thd->fd);
	  client->thd->fd = -1;
	  return -1;
	}
    }
  while (ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN);

  if (ret < 0)
    {
      log_write ("%s", gnutls_strerror (ret));
      ret = 0;
    }

  return ret;
}

ssize_t
tls_write_hook (assuan_context_t ctx, assuan_fd_t fd, const void *data,
		size_t len)
{
  struct client_s *client = assuan_get_pointer (ctx);
  ssize_t ret;
  time_t start = time (NULL);

  do
    {
      gpg_error_t rc = 0;

      ret = gnutls_record_send (client->thd->tls->ses, data, len);
      TEST_TIMEOUT (client->thd->timeout, ret, start, rc);
      if (rc)
	{
	  errno = ETIMEDOUT;
	  close (client->thd->fd);
	  client->thd->fd = -1;
	  return -1;
	}
    }
  while (ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN);

  return ret;
}

gpg_error_t
tls_init_params ()
{
  int n;
  char *tmp, *tmp2;
  gpg_error_t rc = GPG_ERR_UNKNOWN_ERRNO;

  n = gnutls_certificate_allocate_credentials (&x509_cred);
  if (n != GNUTLS_E_SUCCESS)
    {
      log_write ("%s", gnutls_strerror (n));
      x509_cred = NULL;
      goto fail;
    }

  tmp = str_asprintf ("%s/ca-cert.pem", homedir);
  if (!tmp)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  n = gnutls_certificate_set_x509_trust_file (x509_cred, tmp,
					      GNUTLS_X509_FMT_PEM);
  if (n < 0)
    {
      log_write ("%s: %s", tmp, gnutls_strerror (n));
      xfree (tmp);
      goto fail;
    }

  xfree (tmp);
  tmp = str_asprintf ("%s/server-cert.pem", homedir);
  if (!tmp)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  tmp2 = str_asprintf ("%s/server-key.pem", homedir);
  if (!tmp2)
    {
      xfree (tmp);
      log_write ("%s(%i): %s", __FILE__, __LINE__, strerror (ENOMEM));
      goto fail;
    }

  n = gnutls_certificate_set_x509_key_file (x509_cred, tmp, tmp2,
					    GNUTLS_X509_FMT_PEM);
  xfree (tmp);
  xfree (tmp2);
  if (n != GNUTLS_E_SUCCESS)
    {
      log_write ("%s", gnutls_strerror (n));
      goto fail;
    }

  log_write ("%s", _("Generating key exchange parameters..."));
  n = gcry_control (GCRYCTL_ENABLE_QUICK_RANDOM, 0);
  if (n)
    {
      log_write ("%s", pwmd_strerror (n));
      goto fail;
    }

  n = gnutls_dh_params_init (&dh_params);
  if (n != GNUTLS_E_SUCCESS)
    {
      log_write ("%s", gnutls_strerror (n));
      goto fail;
    }

  n = gnutls_dh_params_generate2 (dh_params, 1024);
  if (n != GNUTLS_E_SUCCESS)
    {
      log_write ("%s", gnutls_strerror (n));
      goto fail;
    }

  gnutls_certificate_set_dh_params (x509_cred, dh_params);
  n = gnutls_rsa_params_init (&rsa_params);
  if (n != GNUTLS_E_SUCCESS)
    {
      log_write ("%s", gnutls_strerror (n));
      goto fail;
    }

  n = gnutls_rsa_params_generate2 (rsa_params, 512);
  if (n != GNUTLS_E_SUCCESS)
    {
      log_write ("%s", gnutls_strerror (n));
      goto fail;
    }

  gnutls_certificate_set_rsa_export_params (x509_cred, rsa_params);
  gnutls_certificate_set_params_function (x509_cred, tls_get_params);
  return 0;

fail:
  return rc;
}

void
tls_deinit_params ()
{
  if (dh_params)
    {
      gnutls_dh_params_deinit (dh_params);
      dh_params = NULL;
    }

  if (rsa_params)
    {
      gnutls_rsa_params_deinit (rsa_params);
      rsa_params = NULL;
    }

  if (x509_cred)
    {
      gnutls_certificate_free_credentials (x509_cred);
      x509_cred = NULL;
    }
}

static gpg_error_t
do_tls_validate_access (struct client_s *client, const char *section)
{
  char **access = config_get_list (section, "tls_access");
  char **p;
  int allowed = 0;

  if (!access || !*access)
    return GPG_ERR_INV_USER_ID;

  for (p = access; p && *p; p++)
    {
      int not = 0;
      char *fp = *p;

      if (*fp && *fp == '+' && *(fp + 1) == 0)
	{
	  allowed = 1;
	  continue;
	}

      if (*fp == '!' || *fp == '-')
	{
	  not = 1;
	  fp++;
	  if (!*fp)
	    break;
	}

      if (!strcasecmp (client->thd->tls->fp, fp))
	allowed = not ? 0 : 1;
    }

  strv_free (access);
  return allowed ? 0 : GPG_ERR_INV_USER_ID;
}

gpg_error_t
tls_validate_access (struct client_s *client, const char *filename)
{
  gpg_error_t rc = do_tls_validate_access (client, "global");

  if (!rc && filename)
    rc = do_tls_validate_access (client, filename);

  return rc;
}
