/*
    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
    Ben Kibbey <bjk@luxsci.net>

    This file is part of pwmd.

    Pwmd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    Pwmd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Pwmd.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <libxml/xmlwriter.h>

#ifndef _
#include "gettext.h"
#define _(msgid) gettext(msgid)
#endif

#include "pwmd-error.h"
#include "util-misc.h"
#include "xml.h"
#include "mem.h"
#include "rcfile.h"

extern void log_write (const char *fmt, ...);

/*
 * 'element' must be allocated.
 */
int
is_literal_element (char **element)
{
  char *p;

  if (!element || !*element)
    return 0;

  if (*(*element) == '!')
    {
      char *c;

      for (p = *element, c = p + 1; *c; c++)
	*p++ = *c;

      *p = 0;
      return 1;
    }

  return 0;
}

int
valid_xml_attribute (const char *str)
{
  return valid_xml_element ((xmlChar *)str);
}

int
valid_xml_element (xmlChar *str)
{
  wchar_t *wc;
  size_t len, c;

  if (!str || !*str || *str == '!')
    return 0;

  wc = str_to_wchar ((const char *)str);
  len = wcslen (wc);

  for (c = 0; c < len; c++)
    {
      switch (wc[c])
	{
	case '-':
	case '.':
	case '0' ... '9':
	case 0xB7:
	case 0x0300 ... 0x036F:
	case 0x203F ... 0x2040:
	  if (!c)
	    {
	      xfree (wc);
	      return 0;
	    }
	case ':': break;
	case '_': break;
	case 'A' ... 'Z': break;
	case 'a' ... 'z': break;
	case 0xC0 ... 0xD6: break;
	case 0xD8 ... 0xF6: break;
	case 0xF8 ... 0x2FF: break;
	case 0x370 ... 0x37D: break;
	case 0x37F ... 0x1FFF: break;
	case 0x200C ... 0x200D: break;
	case 0x2070 ... 0x218F: break;
	case 0x2C00 ... 0x2FEF: break;
	case 0x3001 ... 0xD7FF: break;
	case 0xF900 ... 0xFDCF: break;
	case 0xFDF0 ... 0xFFFD: break;
	case 0x10000 ... 0xEFFFF: break;
	default:
	  xfree (wc);
	  return 0;
	}
    }

  xfree (wc);
  return 1;
}

int
valid_element_path (char **path, int with_content)
{
  char **dup = NULL, **p;
  int i, t;

  if (!path || !*path)
    return 0;

  /* Save some memory by not duplicating the element content. */
  if (with_content)
    {
      t = strv_length (path);
      for (i = 0; i < t - 1; i++)
	{
	  char **tmp = xrealloc (dup, (i + 2) * sizeof (char *));

	  if (!tmp)
	    {
	      strv_free (dup);
	      return 0;
	    }

	  dup = tmp;
	  dup[i] = str_dup (path[i]);
	  dup[i + 1] = NULL;
	}
    }
  else
    dup = strv_dup (path);

  if (!dup)
    return 0;

  for (p = dup; *p && *(*p); p++)
    {
      is_literal_element (&(*p));
      if (!valid_xml_element ((xmlChar *) * p))
	{
	  strv_free (dup);
	  return 0;
	}
    }

  strv_free (dup);
  return 1;
}

gpg_error_t
attr_ctime (xmlNodePtr n)
{
  char *buf = str_asprintf ("%li", time (NULL));
  gpg_error_t rc;

  if (!buf)
    return GPG_ERR_ENOMEM;

  rc = add_attribute (n, "_ctime", buf);
  xfree (buf);
  return rc;
}

static gpg_error_t
create_new_element (xmlNodePtr parent, const char *name, xmlNodePtr * result)
{
  xmlNodePtr n = xmlNewNode (NULL, (xmlChar *) "element");
  gpg_error_t rc;

  if (!n)
    return GPG_ERR_ENOMEM;

  rc = add_attribute (n, "_name", name);
  if (!rc)
    rc = attr_ctime (n);

  if (!rc)
    {
      if (result)
	*result = xmlAddChild (parent, n);
      else
	(void) xmlAddChild (parent, n);
    }
  else
    xmlFreeNode (n);

  return rc;
}

gpg_error_t
new_root_element (xmlDocPtr doc, char *name)
{
  xmlNodePtr root = xmlDocGetRootElement (doc);
  char *p = name;

  if (!p || !root)
    return GPG_ERR_BAD_DATA;

  if (*p == '!')
    p++;

  if (!valid_xml_element ((xmlChar *) p))
    return GPG_ERR_INV_VALUE;

  return create_new_element (root, p, NULL);
}

static xmlDocPtr
create_dtd ()
{
  xmlDocPtr doc;
  xmlTextWriterPtr wr = xmlNewTextWriterDoc (&doc, 0);

  if (!wr)
    return NULL;

  if (xmlTextWriterStartDocument (wr, NULL, "UTF-8", "yes"))
    goto fail;

  if (xmlTextWriterStartDTD (wr, (xmlChar *) "pwmd", NULL, NULL) == -1)
    goto fail;

  if (xmlTextWriterWriteDTDElement (wr, (xmlChar *) "pwmd",
				    (xmlChar *) "(element)") == -1)
    goto fail;

  xmlTextWriterEndDTDElement (wr);

  if (xmlTextWriterWriteDTDAttlist (wr, (xmlChar *) "element",
				    (xmlChar *) "_name CDATA #REQUIRED") ==
      -1)
    goto fail;

  xmlTextWriterEndDTDAttlist (wr);
  xmlTextWriterEndDTD (wr);

  if (xmlTextWriterStartElement (wr, (xmlChar *) "pwmd"))
    goto fail;

  xmlTextWriterEndElement (wr);
  xmlTextWriterEndDocument (wr);
  xmlFreeTextWriter (wr);
  return doc;

fail:
  xmlTextWriterEndDocument (wr);
  xmlFreeTextWriter (wr);
  xmlFreeDoc (doc);
  return NULL;
}

xmlDocPtr
new_document ()
{
  return create_dtd ();
}

xmlNodePtr
find_element_node (xmlNodePtr node)
{
  xmlNodePtr n = node;

  if (n && n->type == XML_ELEMENT_NODE)
    return n;

  for (n = node; n; n = n->next)
    {
      if (n->type == XML_ELEMENT_NODE)
	return n;
    }

  return NULL;
}

static xmlNodePtr
resolve_path (xmlDocPtr doc, xmlChar * path, char ***result, gpg_error_t * rc)
{
  xmlNodePtr n;
  char **req;

  req = str_split ((char *) path, "\t", 0);
  if (!req)
    {
      *rc = GPG_ERR_ENOMEM;
      return NULL;
    }

  n = find_root_element (doc, &req, rc, NULL, 0, 0);
  if (!n)
    {
      strv_free (req);
      return NULL;
    }

  if (req[1])
    n = find_elements (doc, n->children, req + 1, rc, NULL, NULL, NULL,
		       0, 0, NULL, 0);

  if (*rc)
    strv_free (req);
  else
    *result = req;

  return n;
}

/*
 * Lists root element names; the value of the attribute "_name" of an element
 * "element". If there's a target attribute both literal and non-literal
 * element names will be added. This is the primary reason why XML entities
 * cannot be used. There wouldn't be a way to get the literal an non-literal
 * element paths.
 */
gpg_error_t
list_root_elements (xmlDocPtr doc, struct string_s ** result,
		    int verbose, int with_target)
{
  xmlNodePtr n = NULL;
  struct slist_s *list = NULL;
  int total, i;
  struct string_s *string;
  gpg_error_t rc = 0;

  n = xmlDocGetRootElement (doc);
  if (!n || !n->children)
    return GPG_ERR_NO_DATA;

  for (n = n->children; n; n = n->next)
    {
      xmlAttrPtr a;
      xmlChar *val, *target;
      struct slist_s *tlist;
      char *tmp;

      if (n->type != XML_ELEMENT_NODE)
	continue;

      a = xmlHasProp (n, (xmlChar *) "_name");
      if (!a || !a->children->content)
	continue;

      val = xmlNodeGetContent (a->children);
      if (!val)
	{
	  rc = GPG_ERR_ENOMEM;
	  goto fail;
	}

      tmp =
	str_asprintf ("!%s%s", (char *) val,
		      verbose ? find_element_node (n->children) ? " +" : "" :
		      "");

      if (!tmp)
	{
	  xmlFree (val);
	  rc = GPG_ERR_ENOMEM;
	  goto fail;
	}

      tlist = slist_append (list, tmp);
      if (!tlist)
	{
	  xmlFree (val);
	  rc = GPG_ERR_ENOMEM;
	  goto fail;
	}

      list = tlist;
      target = node_has_attribute (n, (xmlChar *) "target");
      if (target)
	{
	  char *t = NULL;

	  if (verbose)
	    {
	      char **req = NULL;
	      xmlNodePtr tnode = resolve_path (doc, target, &req, &rc);

	      if (rc == GPG_ERR_ELEMENT_NOT_FOUND || rc == GPG_ERR_ELOOP)
		{
		  t = str_asprintf ("%s %s", (char *) val,
				    rc == GPG_ERR_ELOOP ? "O" : "E");
		  rc = 0;
		}
	      else if (!rc)
		{
		  struct string_s *realpath = NULL;

		  if (with_target)
		    {
		      rc = build_realpath (doc, (char *) target, &realpath);
		      if (rc)
			{
			  strv_free (req);
			  xmlFree (val);
			  xmlFree (target);
			  goto fail;
			}

		      realpath = string_prepend (realpath, "T ");
		    }

		  t = str_asprintf ("%s%s%s%s", (char *) val,
				    (tnode
				     && find_element_node (tnode->children))
				    || realpath ? " " : "", tnode
				    && find_element_node (tnode->children) ?
				    "+" : "", realpath ? realpath->str : "");

		  if (realpath)
		    string_free (realpath, 1);
		}

	      if (req)
		strv_free (req);
	    }
	  else
	    t = str_dup ((char *) val);

	  if (!t || rc)
	    {
	      xmlFree (val);
	      xmlFree (target);
	      rc = rc ? rc : GPG_ERR_ENOMEM;
	      goto fail;
	    }

	  tlist = slist_append (list, t);
	  if (!tlist)
	    {
	      xmlFree (val);
	      xfree (t);
	      xmlFree (target);
	      rc = GPG_ERR_ENOMEM;
	      goto fail;
	    }

	  list = tlist;
	}

      xmlFree (val);
      xmlFree (target);
    }

  total = slist_length (list);
  if (!total)
    return GPG_ERR_NO_DATA;

  string = string_new (NULL);
  if (!string)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  for (i = 0; i < total; i++)
    {
      char *val = slist_nth_data (list, i);

      string_append_printf (string, "%s\n", val);
    }

  string = string_truncate (string, string->len - 1);
  *result = string;

fail:
  total = slist_length (list);
  for (i = 0; i < total; i++)
    xfree (slist_nth_data (list, i));

  slist_free (list);
  return rc;
}

/*
 * Prevents a sibling element past the current element path with the same
 * element name.
 */
static xmlNodePtr
find_stop_node (xmlNodePtr node)
{
  xmlNodePtr n;

  for (n = node->parent->children; n; n = n->next)
    {
      if (n == node)
	return n->next;
    }

  return NULL;
}

xmlNodePtr
create_target_elements_cb (xmlNodePtr node, char **path,
			   gpg_error_t * rc, void *data)
{
  int i;
  char **req = path;
  xmlNodePtr parent = data;

  for (i = 0; req[i] && *req[i]; i++)
    {
      xmlNodePtr n;

      if (parent && node == parent)
	{
	  *rc = GPG_ERR_CONFLICT;
	  return NULL;
	}

      is_literal_element (&req[i]);

      if ((n = find_element (node, req[i], find_stop_node (node))) == NULL ||
	  (n && n->parent == node->parent))
	{

	  *rc = create_new_element (node, req[i], &node);
	  if (*rc)
	    return NULL;
	}
      else
	node = n;
    }

  return node;
}

xmlNodePtr
find_text_node (xmlNodePtr node)
{
  xmlNodePtr n = node;

  if (n && n->type == XML_TEXT_NODE)
    return n;

  for (n = node; n; n = n->next)
    {
      if (n->type == XML_TEXT_NODE)
	return n;
    }

  return NULL;
}

xmlNodePtr
create_elements_cb (xmlNodePtr node, char **elements,
		    gpg_error_t * rc, void *data)
{
  int i;
  char **req = elements;

  if (node->type == XML_TEXT_NODE)
    node = node->parent;

  for (i = 0; req[i] && *req[i]; i++)
    {
      xmlNodePtr n;

      /*
       * Strip the first '!' if needed. If there's another, it's an
       * rc. The syntax has already been checked before calling this
       * function.
       */
      is_literal_element (&req[i]);
      n = find_element (node, req[i], find_stop_node (node));

      /*
       * If the found element has the same parent as the current element,
       * they are siblings and the new element needs to be created as a
       * child of the current element (node).
       */
      if (n && n->parent == node->parent)
	n = NULL;

      if (!n)
	{
	  *rc = create_new_element (node, req[i], &node);
	  if (*rc)
	    return NULL;
	}
      else
	node = n;
    }

  return node;
}

/* The root element is really req[0]. It is need as a pointer in case there is
 * a target attribute so it can be updated. */
xmlNodePtr
find_root_element (xmlDocPtr doc, char ***req, gpg_error_t * rc,
		   int *target, int recursion_depth, int stop)
{
  xmlNodePtr n = xmlDocGetRootElement (doc);
  int depth = 0;
  char *root = str_dup (*req[0]);
  int literal = is_literal_element (&root);

  if (!root)
    {
      *rc = GPG_ERR_ENOMEM;
      return NULL;
    }

  *rc = 0;
  recursion_depth++;

  if (max_recursion_depth >= 1 && recursion_depth > max_recursion_depth)
    {
      xmlChar *t = xmlGetNodePath (n);

      log_write ("%s: %s", pwmd_strerror (GPG_ERR_ELOOP), t);
      xmlFree (t);
      xfree (root);
      *rc = GPG_ERR_ELOOP;
      return NULL;
    }

  while (n)
    {
      if (n->type == XML_ELEMENT_NODE)
	{
	  if (depth == 0 && xmlStrEqual (n->name, (xmlChar *) "pwmd"))
	    {
	      n = n->children;
	      depth++;
	      continue;
	    }

	  if (depth == 1 && xmlStrEqual (n->name, (xmlChar *) "element"))
	    {
	      xmlChar *content = node_has_attribute (n, (xmlChar *) "_name");

	      if (!content)
		continue;

	      if (xmlStrEqual (content, (xmlChar *) root))
		{
		  char **nreq, **tmp = NULL;

		  if (literal == 1)
		    {
		      xmlFree (content);
		      xfree (root);
		      return n;
		    }

		  xmlFree (content);
		  content = node_has_attribute (n, (xmlChar *) "target");

		  if (target)
		    *target = 1;

		  if (!content || stop)
		    {
		      if (content)
			xmlFree (content);

		      xfree (root);
		      return n;
		    }

		  if (strchr ((char *) content, '\t'))
		    {
		      nreq = str_split ((char *) content, "\t", 0);
		      xmlFree (content);

#if 0
		      /*
		       * FIXME ENOMEM
		       */
		      if (!nreq)
			{
			  *rc = GPG_ERR_ENOMEM;
			  return NULL;
			}
#endif

		      tmp = *req;
		      tmp = strv_catv (nreq, tmp + 1);
		      strv_free (nreq);

		      if (!tmp)
			{
			  xfree (root);
			  *rc = GPG_ERR_ENOMEM;
			  return NULL;
			}

		      strv_free (*req);
		      *req = tmp;
		    }
		  else
		    {
		      if (strv_printf (&tmp, "%s", content) == 0)
			{
			  xmlFree (content);
			  xfree (root);
			  *rc = GPG_ERR_ENOMEM;
			  return NULL;
			}

		      xmlFree (content);
		      nreq = *req;
		      nreq = strv_catv (tmp, nreq + 1);
		      strv_free (tmp);

		      if (!nreq)
			{
			  *rc = GPG_ERR_ENOMEM;
			  xfree (root);
			  return NULL;
			}

		      strv_free (*req);
		      *req = nreq;
		    }

		  xfree (root);
		  n =
		    find_root_element (doc, req, rc, target, recursion_depth,
				       0);
		  return n;
		}

	      xmlFree (content);
	    }
	}

      n = n->next;
    }

  xfree (root);
  *rc = GPG_ERR_ELEMENT_NOT_FOUND;
  return NULL;
}

xmlNodePtr
find_element (xmlNodePtr node, char *element, xmlNodePtr stop)
{
  xmlNodePtr n;

  if (!node || !element)
    return NULL;

  for (n = node; n; n = n->next)
    {
      if (n->type != XML_ELEMENT_NODE)
	continue;

      if (n == stop)
	break;

      xmlChar *a = node_has_attribute (n, (xmlChar *) "_name");

      if (a && xmlStrEqual (a, (xmlChar *) element))
	{
	  xmlFree (a);
	  return n;
	}

      xmlFree (a);
    }

  return NULL;
}

xmlChar *
node_has_attribute (xmlNodePtr n, xmlChar * attr)
{
  xmlAttrPtr a = xmlHasProp (n, attr);

  if (!a)
    return NULL;

  if (!a->children || !a->children->content)
    return NULL;

  return xmlGetProp (n, attr);
}

static int
element_to_literal (char **element)
{
  char *p = str_asprintf ("!%s", *element);

  if (!p)
    return 0;

  xfree (*element);
  *element = p;
  return 1;
}

/* Resolves elements in 'req' one at a time. It's recursive in case of
 * "target" attributes. */
xmlNodePtr
find_elements (xmlDocPtr doc, xmlNodePtr node,
	       char **req, gpg_error_t * rc, int *target,
	       xmlNodePtr (*found_fn) (xmlNodePtr, char **, gpg_error_t *,
				       char **, void *),
	       xmlNodePtr (*not_found_fn) (xmlNodePtr, char **, gpg_error_t *,
					   void *), int is_list_command,
	       int recursion_depth, void *data, int stop)
{
  xmlNodePtr n, last, last_node;
  char **p;
  int found = 0;

  *rc = 0;
  recursion_depth++;

  if (max_recursion_depth >= 1 && recursion_depth > max_recursion_depth)
    {
      xmlChar *t = xmlGetNodePath (node);

      log_write ("%s: %s", pwmd_strerror (GPG_ERR_ELOOP), t);
      xmlFree (t);
      recursion_depth--;
      *rc = GPG_ERR_ELOOP;
      return NULL;
    }

  for (last_node = last = n = node, p = req; *p; p++)
    {
      xmlNodePtr tmp;
      char *t;
      int literal;

      if (!*(*p))
	{
	  *rc = GPG_ERR_ELEMENT_NOT_FOUND;
	  return NULL;
	}

      t = str_dup (*p);
      if (!t)
	{
	  *rc = GPG_ERR_ENOMEM;
	  return NULL;
	}

      literal = is_literal_element (&t);
      n = find_element (last, t, NULL);
      xfree (t);

      if (!n)
	{
	  if (not_found_fn)
	    return not_found_fn (found ? last_node : last_node->parent, p, rc,
				 data);

	  *rc = GPG_ERR_ELEMENT_NOT_FOUND;
	  return NULL;
	}

      last = n->children;
      last_node = n;
      found = 1;

      if (literal == 0)
	{
	  xmlChar *content = node_has_attribute (n, (xmlChar *) "target");
	  char **nreq = NULL, **nnreq;

	  if (!content)
	    {
	      if (is_list_command == 1)
		{
		  if (element_to_literal (&(*p)) == 0)
		    {
		      *rc = GPG_ERR_ENOMEM;
		      return NULL;
		    }
		}

	      continue;
	    }

	  if (target)
	    *target = 1;

	  if (!*(p + 1) && stop)
	    {
	      xmlFree (content);
	      return n;
	    }

	  if (strchr ((char *) content, '\t') != NULL)
	    {
	      if ((nreq = str_split ((char *) content, "\t", 0)) == NULL)
		{
		  xmlFree (content);
		  *rc = GPG_ERR_INV_VALUE;
		  return NULL;
		}
	    }
	  else
	    {
	      if ((nreq = str_split ((char *) content, " ", 0)) == NULL)
		{
		  xmlFree (content);
		  *rc = GPG_ERR_INV_VALUE;
		  return NULL;
		}
	    }

	  xmlFree (content);
	  tmp = find_root_element (doc, &nreq, rc, target, 0, 0);

	  if (!tmp)
	    {
	      strv_free (nreq);
	      return NULL;
	    }

	  if (found_fn)
	    {
	      found_fn (tmp, nreq, rc, p + 1, data);

	      if (*rc)
		{
		  strv_free (nreq);
		  return NULL;
		}
	    }

	  if (!*(nreq + 1) && !*(p + 1))
	    {
	      strv_free (nreq);
	      return tmp;
	    }

	  nnreq = strv_catv (nreq + 1, p + 1);
	  strv_free (nreq);

	  // FIXME ENOMEM
	  if (!nnreq || !*nnreq)
	    {
	      if (nnreq)
		strv_free (nnreq);

	      return tmp;
	    }

	  if (tmp->children)
	    n = find_elements (doc, tmp->children, nnreq, rc, NULL, found_fn,
			       not_found_fn, is_list_command, recursion_depth,
			       data, stop);
	  else
	    {
	      strv_free (nnreq);

	      if (not_found_fn)
		return not_found_fn (tmp, p + 1, rc, data);

	      *rc = GPG_ERR_ELEMENT_NOT_FOUND;
	      return NULL;
	    }

	  if (*(p + 1))
	    {
	      char **zz = p + 1, **qq = nnreq;

	      if (strv_length (nnreq) > strv_length (p + 1))
		qq = nnreq + 1;

	      for (; *qq && *zz; zz++)
		{
		  xfree (*zz);
		  *zz = str_dup (*qq++);

		  if (!*zz)
		    {
		      *rc = GPG_ERR_ENOMEM;
		      n = NULL;
		      break;
		    }
		}
	    }

	  strv_free (nnreq);
	  return n;
	}
    }

  return n;
}

static int
update_element_list (struct element_list_s *elements)
{
  char *line;
  struct slist_s *l;

  if (!elements || !elements->elements)
    return 1;

  line = strv_join ("\t", elements->elements);

  if (!line)
    return 0;

  strv_free (elements->elements);
  elements->elements = NULL;
  l = slist_append (elements->list, line);

  if (!l)
    return 0;

  elements->list = l;
  return 1;
}

static gpg_error_t
path_list_recurse (xmlDocPtr doc, xmlNodePtr node,
		   struct element_list_s *elements)
{
  gpg_error_t rc = 0;
  xmlNodePtr n;
  gpg_error_t error_flag = 0;

  for (n = node; n; n = n->next)
    {
      xmlChar *target = NULL;
      xmlChar *a = node_has_attribute (n, (xmlChar *) "_name");
      int err = 0;

      if (!a)
	continue;

      if (n->type != XML_ELEMENT_NODE)
	goto children;

      if (elements->verbose)
	{
	  if (strv_printf
	      (&elements->elements, "%s\t!%s%s", elements->prefix, a,
	       find_element_node (n->children) ? " +" : "") == 0)
	    {
	      xmlFree (a);
	      return GPG_ERR_ENOMEM;
	    }
	}
      else
	if (strv_printf (&elements->elements, "%s\t!%s", elements->prefix, a)
	    == 0)
	{
	  xmlFree (a);
	  return GPG_ERR_ENOMEM;
	}

      if (update_element_list (elements) == 0)
	{
	  xmlFree (a);
	  return GPG_ERR_ENOMEM;
	}

      target = node_has_attribute (n, (xmlChar *) "target");

      if (target)
	{
	  char *tmp;
	  char *save = elements->prefix;
	  int r = elements->resolving;
	  char **req = NULL;
	  xmlNodePtr tnode;
	  struct string_s *realpath = NULL;

	  tnode = resolve_path (doc, target, &req, &rc);
	  if (rc == GPG_ERR_ELOOP || rc == GPG_ERR_ELEMENT_NOT_FOUND)
	    {
	      if (rc == GPG_ERR_ELOOP)
		{
		  xmlChar *t = xmlGetNodePath (n);

		  log_write ("%s: %s", pwmd_strerror (GPG_ERR_ELOOP), t);
		  xmlFree (t);
		}

	      if (elements->verbose)
		{
		  strv_printf (&elements->elements, "%s\t%s %s",
			       elements->prefix, a,
			       rc == GPG_ERR_ELOOP ? "O" : "E");
		  error_flag = err = rc;
		  rc = 0;
		  goto update;
		}
	    }

	  if (!elements->verbose && rc)
	    {
	      xmlFree (a);
	      xmlFree (target);
	      return rc;
	    }

	  if (elements->with_target)
	    {
	      rc = build_realpath (doc, (char *) target, &realpath);
	      if (rc)
		{
		  xmlFree (a);
		  xmlFree (target);
		  return rc;
		}

	      realpath = string_prepend (realpath, "T ");
	    }

	  if (elements->verbose)
	    {
	      if (!strv_printf (&elements->elements, "%s\t%s%s%s%s",
				elements->prefix, a,
				(tnode && find_element_node (tnode->children))
				|| realpath ? " " : "", tnode
				&& find_element_node (tnode->children) ? "+" :
				"", realpath ? realpath->str : ""))
		{
		  xmlFree (a);
		  xmlFree (target);
		  return GPG_ERR_ENOMEM;
		}
	    }
	  else
	    if (!strv_printf
		(&elements->elements, "%s\t%s", elements->prefix, a))
	    {
	      xmlFree (a);
	      xmlFree (target);
	      return GPG_ERR_ENOMEM;
	    }

	update:
	  if (realpath)
	    string_free (realpath, 1);

	  tmp = strv_join ("\t", elements->elements);

	  if (!tmp)
	    {
	      xmlFree (a);
	      xmlFree (target);
	      return GPG_ERR_ENOMEM;
	    }

	  if (update_element_list (elements) == 0)
	    {
	      xfree (tmp);
	      xmlFree (a);
	      xmlFree (target);
	      return GPG_ERR_ENOMEM;
	    }

	  if (!err && elements->recurse)
	    {
	      /* Prune element flags. */
	      if (elements->verbose && strchr (tmp, ' '))
		{
		  char *p;

		  for (p = tmp; *p; p++)
		    {
		      if (*p == ' ')
			{
			  *p = 0;
			  break;
			}
		    }
		}

	      elements->prefix = tmp;
	      elements->resolving = 1;
	      rc = create_path_list (doc, elements, (char *) target);
	      elements->resolving = r;
	      elements->prefix = save;

	      if (rc && gpg_err_code (rc) != GPG_ERR_ELOOP)
		{
		  xfree (tmp);
		  xmlFree (target);
		  xmlFree (a);
		  return rc;
		}

	      error_flag = err = rc;
	      rc = 0;
	    }

	  xfree (tmp);
	  xmlFree (target);
	}

    children:
      if (n->children && elements->recurse && err != GPG_ERR_ELOOP)
	{
	  char *tmp = str_asprintf ("%s\t!%s", elements->prefix, a);
	  char *save = elements->prefix;

	  if (!tmp)
	    {
	      xmlFree (a);
	      return GPG_ERR_ENOMEM;
	    }

	  elements->prefix = tmp;
	  rc = path_list_recurse (doc, n->children, elements);
	  xfree (elements->prefix);
	  elements->prefix = save;

	  if (rc)
	    {
	      xmlFree (a);
	      return rc;
	    }
	}

      xmlFree (a);
    }

  return error_flag == GPG_ERR_ELOOP ? error_flag : rc;
}

gpg_error_t
add_attribute (xmlNodePtr node, const char *name, const char *value)
{
  char *buf;
  gpg_error_t rc;

  if (name && !xmlSetProp (node, (xmlChar *) name, (xmlChar *) value))
    return GPG_ERR_BAD_DATA;

  if (name && xmlStrEqual ((xmlChar *) name, (xmlChar *) "_mtime"))
    return 0;

  buf = str_asprintf ("%li", time (NULL));
  rc = add_attribute (node, "_mtime", buf);
  xfree (buf);
  return rc;
}

/*
 * From the element path 'path', find sub-nodes and append them to the list.
 */
gpg_error_t
create_path_list (xmlDocPtr doc, struct element_list_s * elements, char *path)
{
  gpg_error_t rc;
  char **req, **req_orig;
  xmlNodePtr n;
  int a_target = 0;

  req = str_split (path, "\t", 0);
  if (!req)
    {
      req = str_split (path, " ", 0);
      if (!req)
	return GPG_ERR_SYNTAX;
    }

  req_orig = strv_dup (req);
  if (!req_orig)
    {
      rc = GPG_ERR_ENOMEM;
      goto fail;
    }

  n = find_root_element (doc, &req, &rc, &a_target, 0, 0);
  if ((rc == GPG_ERR_ELEMENT_NOT_FOUND || rc == GPG_ERR_ELOOP)
      && elements->verbose && a_target)
    {
      rc = 0;
      goto done;
    }

  if (!n && rc == GPG_ERR_ELEMENT_NOT_FOUND && elements->resolving == 1)
    {
      rc = 0;
      goto fail;
    }
  else if (!n)
    goto fail;

  if (a_target == 1)
    {
      xfree (*req);
      *req = str_dup (*req_orig);
    }

  if (*(req + 1))
    {
      int e_target = 0;

      n =
	find_elements (doc, n->children, req + 1, &rc, &e_target, NULL, NULL,
		       1, 0, NULL, 0);

      if (!n && rc == GPG_ERR_ELEMENT_NOT_FOUND && elements->resolving == 1)
	{
	  rc = 0;
	  goto fail;
	}
      else if (!n)
	goto fail;
    }

done:
  if (!elements->prefix)
    {
      /*
       * FIXME
       *
       * If any req_orig element contains no target the element should be
       * prefixed with the literal character. Not really crucial if the
       * client isn't human because child elements are prefixed for the
       * current path. But may be confusing if editing by hand.
       */
      elements->prefix = strv_join ("\t", req_orig);

      if (!elements->prefix)
	{
	  rc = GPG_ERR_ENOMEM;
	  goto fail;
	}

      if (elements->verbose)
	{
	  int ret;
	  struct string_s *realpath = NULL;

	  if (a_target && elements->with_target)
	    {
	      rc = build_realpath (doc, path, &realpath);
	      if (rc)
		goto fail;

	      realpath = string_prepend (realpath, "T ");
	    }

	  ret = strv_printf (&elements->elements, "%s%s%s%s",
			     elements->prefix,
			     find_element_node (n->children)
			     || realpath ? " " : "",
			     find_element_node (n->children) ? "+" : "",
			     realpath ? realpath->str : "");
	  string_free (realpath, 1);
	  if (!ret)
	    {
	      rc = GPG_ERR_ENOMEM;
	      goto fail;
	    }
	}
      else if (strv_printf (&elements->elements, "%s", elements->prefix) == 0)
	{
	  rc = GPG_ERR_ENOMEM;
	  goto fail;
	}

      if (update_element_list (elements) == 0)
	{
	  rc = GPG_ERR_ENOMEM;
	  goto fail;
	}
    }

  rc = path_list_recurse (doc, n->children, elements);

fail:
  if (req_orig)
    strv_free (req_orig);

  strv_free (req);
  return rc;
}

gpg_error_t
recurse_xpath_nodeset (xmlDocPtr doc, xmlNodeSetPtr nodes,
		       xmlChar * value, xmlBufferPtr * result, int cmd,
		       const xmlChar * attr)
{
  int i = value ? nodes->nodeNr - 1 : 0;
  xmlBufferPtr buf;

  buf = xmlBufferCreate ();

  if (!buf)
    return GPG_ERR_ENOMEM;

  for (; value ? i >= 0 : i < nodes->nodeNr; value ? i-- : i++)
    {
      xmlNodePtr n = nodes->nodeTab[i];
      gpg_error_t rc;

      if (!n)
	continue;

      if (!value && !attr)
	{
	  if (xmlNodeDump (buf, doc, n, 0, 0) == -1)
	    {
	      *result = buf;
	      return GPG_ERR_BAD_DATA;
	    }

	  continue;
	}

      if (!attr)
	{
	  xmlNodeSetContent (n, value);
	  rc = update_element_mtime (n);

	  if (rc)
	    return rc;
	}
      else
	{
	  if (!cmd)
	    rc = add_attribute (n, (char *) attr, (char *) value);
	  else
	    rc = delete_attribute (n, attr);

	  if (rc)
	    return rc;
	}
    }

  *result = buf;
  return 0;
}

static gpg_error_t
convert_root_element (xmlNodePtr n)
{
  xmlChar *a = xmlGetProp (n, (xmlChar *) "_name");
  gpg_error_t rc;

  if (a)
    {
      xmlFree (a);
      xmlChar *t = xmlGetNodePath (n);

      log_write (_
		 ("An existing \"_name\" attribute already exists. Please rename this attribute before converting. Path is: %s"),
		 t);
      xmlFree (t);
      return GPG_ERR_AMBIGUOUS_NAME;
    }

  a = xmlGetProp (n, (xmlChar *) "name");

  if (a)
    {
      rc = add_attribute (n, "_name", (char *) a);
      xmlFree (a);

      if (rc)
	return rc;

      rc = delete_attribute (n, (xmlChar *) "name");

      if (rc)
	return rc;

      xmlNodeSetName (n, (xmlChar *) "element");
    }

  return 0;
}

gpg_error_t
delete_attribute (xmlNodePtr n, const xmlChar * name)
{
  xmlAttrPtr a;

  if ((a = xmlHasProp (n, name)) == NULL)
    return GPG_ERR_NOT_FOUND;

  if (xmlRemoveProp (a) == -1)
    return GPG_ERR_BAD_DATA;

  return update_element_mtime (n);
}

static gpg_error_t
convert_elements_recurse (xmlDocPtr doc, xmlNodePtr n, unsigned depth)
{
  gpg_error_t rc;

  depth++;

  for (n = n->children; n; n = n->next)
    {
      if (n->type == XML_ELEMENT_NODE)
	{
	  xmlChar *a = NULL;

	  if (depth > 1)
	    {
	      if (xmlStrEqual (n->name, (xmlChar *) "element"))
		{
		  xmlChar *t = xmlGetNodePath (n);

		  log_write (_
			     ("An existing \"element\" already exists. Please rename this element before converting. Path is: %s"),
			     t);
		  xmlFree (t);
		  return GPG_ERR_AMBIGUOUS_NAME;
		}

	      a = xmlGetProp (n, (xmlChar *) "_name");

	      if (a)
		{
		  xmlFree (a);
		  xmlChar *t = xmlGetNodePath (n);

		  log_write (_
			     ("An existing \"_name\" attribute already exists. Please rename this attribute before converting. Path is: %s"),
			     t);
		  xmlFree (t);
		  return GPG_ERR_AMBIGUOUS_NAME;
		}

	      xmlChar *tmp = xmlStrdup (n->name);

	      if (!tmp)
		return GPG_ERR_ENOMEM;

	      xmlNodeSetName (n, (xmlChar *) "element");
	      rc = add_attribute (n, "_name", (char *) tmp);
	      xmlFree (tmp);

	      if (rc)
		return rc;
	    }
	  else
	    {
	      rc = convert_root_element (n);

	      if (rc)
		return rc;
	    }
	}

      if (n->children)
	{
	  rc = convert_elements_recurse (doc, n, depth);

	  if (rc)
	    return rc;
	}
    }

  return 0;
}

/* Renames ALL elements to the new "element" name. Existing element names are
 * stored as an attribute "_name". This was introduced in pwmd 2.12 so
 * elements can contain common characters that the XML parser barfs on (an
 * email address for example. */
gpg_error_t
convert_pre_212_elements (xmlDocPtr doc)
{
  xmlNodePtr n = xmlDocGetRootElement (doc);

  log_write (_("Converting pre 2.12 data file..."));
  return convert_elements_recurse (doc, n, 0);
}

gpg_error_t
validate_import (xmlNodePtr node)
{
  gpg_error_t rc = 0;

  if (!node)
    return 0;

  for (xmlNodePtr n = node; n; n = n->next)
    {
      if (n->type == XML_ELEMENT_NODE)
	{
	  if (xmlStrEqual (n->name, (xmlChar *) "element"))
	    {
	      xmlChar *a = xmlGetProp (n, (xmlChar *) "_name");

	      if (!a)
		{
		  xmlChar *t = xmlGetNodePath (n);

		  log_write (_("Missing attribute '_name' at %s."), t);
		  xmlFree (t);
		  return GPG_ERR_INV_VALUE;
		}

	      if (!valid_xml_element (a))
		{
		  xmlChar *t = xmlGetNodePath (n);

		  log_write (_("'%s' is not a valid element name at %s."), a,
			     t);
		  xmlFree (a);
		  xmlFree (t);
		  return GPG_ERR_INV_VALUE;
		}

	      xmlFree (a);
	      a = xmlGetProp (n, (xmlChar *) "_ctime");
	      if (!a)
		attr_ctime (n);

	      xmlFree (a);
	      a = xmlGetProp (n, (xmlChar *) "_mtime");
	      if (!a)
		update_element_mtime (n);
	      xmlFree (a);
	    }
	  else
	    {
	      xmlChar *t = xmlGetNodePath (n);

	      log_write (_("Warning: unknown element '%s' at %s. Ignoring."),
			 n->name, t);
	      xmlFree (t);
	      continue;
	    }
	}

      if (n->children)
	{
	  rc = validate_import (n->children);

	  if (rc)
	    return rc;
	}
    }

  return rc;
}

gpg_error_t
update_element_mtime (xmlNodePtr n)
{
  return add_attribute (n, NULL, NULL);
}

gpg_error_t
unlink_node (xmlNodePtr n)
{
  gpg_error_t rc = 0;

  if (!n)
    return rc;

  if (n->parent)
    rc = update_element_mtime (n->parent);

  xmlUnlinkNode (n);
  return rc;
}

gpg_error_t
parse_doc (const char *xml, size_t len, xmlDocPtr *result)
{
  xmlDocPtr doc;

  xmlResetLastError ();
  doc = xmlReadMemory (xml, len, NULL, "UTF-8", XML_PARSE_NOBLANKS);
  if (!doc && xmlGetLastError ())
    return GPG_ERR_BAD_DATA;

  *result = doc;
  return !doc ? GPG_ERR_ENOMEM : 0;
}

static xmlNodePtr
realpath_elements_cb (xmlNodePtr node, char **target,
		      gpg_error_t * rc, char **req_orig, void *data)
{
  char *path = *(char **) data;
  char *tmp = NULL, *result;

  if (path)
    {
      xfree (path);
      *(char **) data = NULL;
    }

  path = strv_join ("\t", target);

  if (!path)
    {
      *rc = GPG_ERR_ENOMEM;
      return NULL;
    }

  if (req_orig)
    {
      tmp = strv_join ("\t", req_orig);

      if (!tmp)
	{
	  xfree (path);
	  *rc = GPG_ERR_ENOMEM;
	  return NULL;
	}
    }

  if (tmp && *tmp)
    result = str_asprintf ("%s\t%s", path, tmp);
  else
    result = str_dup (path);

  if (!result)
    {
      *rc = GPG_ERR_ENOMEM;
      xfree (path);
      xfree (tmp);
      return NULL;
    }

  xfree (path);
  xfree (tmp);
  *(char **) data = result;
  return node;
}

gpg_error_t
build_realpath (xmlDocPtr doc, char *line, struct string_s ** result)
{
  gpg_error_t rc;
  char **req;
  char *t;
  int i;
  xmlNodePtr n;
  struct string_s *string;
  char *rp = NULL;

  if (strchr (line, '\t') != NULL)
    {
      if ((req = str_split (line, "\t", 0)) == NULL)
	return GPG_ERR_SYNTAX;
    }
  else
    {
      if ((req = str_split (line, " ", 0)) == NULL)
	return GPG_ERR_SYNTAX;
    }

  n = find_root_element (doc, &req, &rc, NULL, 0, 0);
  if (!n)
    {
      strv_free (req);
      return rc;
    }

  rp = strv_join ("\t", req);
  if (!rp)
    {
      strv_free (req);
      return GPG_ERR_ENOMEM;
    }

  if (req[1])
    {
      n = find_elements (doc, n->children, req + 1, &rc,
			 NULL, realpath_elements_cb, NULL, 0, 0, &rp, 0);
      if (!n)
	{
	  xfree (rp);
	  strv_free (req);
	  return rc;
	}
    }

  string = string_new (rp);
  xfree (rp);
  strv_free (req);
  if (!string)
    return GPG_ERR_ENOMEM;

again:
  for (i = 0, t = string->str + i; *t; t++, i++)
    {
      if ((!i && *t != '!') || (*t == '\t' && *(t + 1) && *(t + 1) != '!'))
	{
	  struct string_s *s = string_insert_c (string, !i ? i++ : ++i, '!');

	  if (!s)
	    {
	      string_free (string, 1);
	      return GPG_ERR_ENOMEM;
	    }

	  string = s;
	  goto again;
	}
    }

  *result = string;
  return rc;
}

#if 0
static char *
node_to_element_path (xmlNodePtr node)
{
  xmlNodePtr n;
  struct string_s *str = string_new ("");
  char *result;

  for (n = node; n; n = n->parent)
    {
      xmlNodePtr child;

      for (child = n; child; child = child->next)
	{
	  if (child->type != XML_ELEMENT_NODE)
	    continue;

	  xmlChar *name = node_has_attribute (n, (xmlChar *) "_name");
	  if (name)
	    {
	      str = string_prepend (str, (char *) name);
	      xmlFree (name);
	      name = node_has_attribute (n, (xmlChar *) "target");
	      if (name)
		str = string_prepend (str, "\t");
	      else
		str = string_prepend (str, "\t!");
	      xmlFree (name);
	    }
	  break;
	}
    }

  str = string_erase (str, 0, 1);
  result = str->str;
  string_free (str, 0);
  return result;
}
#endif

/*
 * Recurse the element tree beginning at 'node' and find elements who point
 * back to 'src' or 'dst'. Also follows target attributes.
 */
static gpg_error_t
find_child_to_target (xmlDocPtr doc, xmlNodePtr node,
		      xmlNodePtr src, xmlNodePtr dst, unsigned depth,
		      int is_target)
{
  xmlNodePtr n;
  gpg_error_t rc = 0;

  if (max_recursion_depth >= 1 && depth > max_recursion_depth)
    return gpg_error (GPG_ERR_ELOOP);

  for (n = node; n; n = n->next)
    {
      xmlChar *target;

      if (n->type != XML_ELEMENT_NODE)
	continue;

      if (n == src || n == dst)
	return GPG_ERR_ELOOP;

      target = node_has_attribute (n, (xmlChar *) "target");
      if (target)
	{
	  xmlNodePtr tmp;
	  char **result = NULL;

	  tmp = resolve_path (doc, target, &result, &rc);
	  xmlFree (target);
	  strv_free (result);
	  if (!rc)
	    {
	      rc = find_child_to_target (doc, tmp, src, dst, ++depth, 1);
	      depth--;
	    }

	  if (rc && gpg_err_code (rc) != GPG_ERR_ELEMENT_NOT_FOUND)
	    return rc;

	  if (is_target)
	    break;

	  continue;
	}

      if (n->children)
	{
	  rc = find_child_to_target (doc, n->children, src, dst, ++depth,0);
	  depth--;
	  if (rc)
	    return rc;
	}

      if (is_target)
	break;
    }

  return rc;
}

static gpg_error_t
find_child_of_parent (xmlDocPtr doc, xmlNodePtr src, xmlNodePtr dst)
{
  xmlNodePtr n;
  gpg_error_t rc = 0;

  for (n = src; n; n = n->next)
    {
      if (n->type != XML_ELEMENT_NODE)
	continue;

      if (n == dst)
	{
	  rc = GPG_ERR_ELOOP;
	  break;
	}

      rc = find_child_of_parent (doc, n->children, dst);
    }

  return rc;
}

static gpg_error_t
find_parent_of_child (xmlDocPtr doc, xmlNodePtr node, xmlNodePtr dst)
{
  xmlNodePtr n;
  gpg_error_t rc = 0;

  for (n = node; n; n = n->parent)
    {
      if (n->type != XML_ELEMENT_NODE)
	{
	  xmlNodePtr tmp;

	  for (tmp = n->next; tmp; n = n->next)
	    {
	      if (n->type != XML_ELEMENT_NODE)
		continue;

	      if (tmp == dst)
		return GPG_ERR_ELOOP;
	    }
	}

      if (n == dst)
	return GPG_ERR_ELOOP;
    }

  return rc;
}

gpg_error_t
validate_target_attribute (xmlDocPtr doc, const char *src,
			   xmlNodePtr dst_node)
{
  gpg_error_t rc;
  xmlNodePtr src_node;
  char **src_req = NULL;

  src_node = resolve_path (doc, (xmlChar *) src, &src_req, &rc);
  if (rc)
    goto fail;

  /* A destination element is a child of the source element. */
  rc = find_child_of_parent (doc, src_node->children, dst_node);
  if (rc)
    goto fail;

  /* The destination element is a parent of the source element. */
  rc = find_parent_of_child (doc, src_node->parent, dst_node);
  if (rc)
    goto fail;

  /* A destination child element contains a target to the source element. */
  rc = find_child_to_target (doc, dst_node->children, src_node, dst_node, 0, 0);
  if (rc)
    goto fail;

fail:
  strv_free (src_req);
  return rc;
}
